@@ -14,17 +14,22 @@ from litex.soc.interconnect.csr_eventmanager import *
 # MAC SRAM Writer ----------------------------------------------------------------------------------
 
 class LiteEthMACSRAMWriter(Module, AutoCSR):
-    def __init__(self, dw, depth, nslots=2, endianness="big"):
+    def __init__(self, dw, depth, nslots=2, endianness="big", timestamp=None):
         self.sink      = sink = stream.Endpoint(eth_phy_description(dw))
         self.crc_error = Signal()
 
-        slotbits   = max(log2_int(nslots), 1)
-        lengthbits = 32
+        slotbits      = max(log2_int(nslots), 1)
+        lengthbits    = 32
 
         self._slot   = CSRStatus(slotbits)
         self._length = CSRStatus(lengthbits)
+        self._errors  = CSRStatus(32)
 
-        self.errors  = CSRStatus(32)
+        if timestamp is not None:
+            # Timestamp the incoming packets when a Timestamp source is provided
+            # and expose value to a software.
+            timestampbits   = len(timestamp)
+            self._timestamp = CSRStatus(timestampbits)
 
         self.submodules.ev = EventManager()
         self.ev.available  = EventSourceLevel()
@@ -59,22 +64,28 @@ class LiteEthMACSRAMWriter(Module, AutoCSR):
         slot_ce = Signal()
         self.sync += If(slot_ce, slot.eq(slot + 1))
 
+        start   = Signal()
         ongoing = Signal()
 
         # Status FIFO
-        fifo = stream.SyncFIFO([("slot", slotbits), ("length", lengthbits)], nslots)
-        self.submodules += fifo
+        stat_fifo_layout = [("slot", slotbits), ("length", lengthbits)]
+        if timestamp is not None:
+            stat_fifo_layout += [("timestamp", timestampbits)]
+
+        stat_fifo = stream.SyncFIFO(stat_fifo_layout, nslots)
+        self.submodules += stat_fifo
 
         # FSM
         self.submodules.fsm = fsm = FSM(reset_state="IDLE")
         fsm.act("IDLE",
             If(sink.valid,
-                If(fifo.sink.ready,
+                If(stat_fifo.sink.ready,
+                    start.eq(1),
                     ongoing.eq(1),
                     NextValue(counter, counter + inc),
                     NextState("WRITE")
                 ).Else(
-                    NextValue(self.errors.status, self.errors.status + 1),
+                    NextValue(self._errors.status, self._errors.status + 1),
                     NextState("DISCARD_REMAINING")
                 )
             )
@@ -106,21 +117,27 @@ class LiteEthMACSRAMWriter(Module, AutoCSR):
             )
         )
         self.comb += [
-            fifo.sink.slot.eq(slot),
-            fifo.sink.length.eq(counter)
+            stat_fifo.sink.slot.eq(slot),
+            stat_fifo.sink.length.eq(counter)
         ]
+
         fsm.act("TERMINATE",
             NextValue(counter, 0),
             slot_ce.eq(1),
-            fifo.sink.valid.eq(1),
+            stat_fifo.sink.valid.eq(1),
             NextState("IDLE")
         )
+
         self.comb += [
-            fifo.source.ready.eq(self.ev.available.clear),
-            self.ev.available.trigger.eq(fifo.source.valid),
-            self._slot.status.eq(fifo.source.slot),
-            self._length.status.eq(fifo.source.length),
+            stat_fifo.source.ready.eq(self.ev.available.clear),
+            self.ev.available.trigger.eq(stat_fifo.source.valid),
+            self._slot.status.eq(stat_fifo.source.slot),
+            self._length.status.eq(stat_fifo.source.length),
         ]
+        if timestamp is not None:
+            # Latch Timestamp on start of incoming packet.
+            self.sync += If(start, stat_fifo.sink.timestamp.eq(timestamp))
+            self.comb += self._timestamp.status.eq(stat_fifo.source.timestamp)
 
         # Memory
         mems  = [None]*nslots
@@ -145,7 +162,7 @@ class LiteEthMACSRAMWriter(Module, AutoCSR):
 # MAC SRAM Reader ----------------------------------------------------------------------------------
 
 class LiteEthMACSRAMReader(Module, AutoCSR):
-    def __init__(self, dw, depth, nslots=2, endianness="big"):
+    def __init__(self, dw, depth, nslots=2, endianness="big", timestamp=None):
         self.source = source = stream.Endpoint(eth_phy_description(dw))
 
         slotbits        = max(log2_int(nslots), 1)
@@ -158,23 +175,43 @@ class LiteEthMACSRAMReader(Module, AutoCSR):
         self._slot   = CSRStorage(slotbits,   reset_less=True)
         self._length = CSRStorage(lengthbits, reset_less=True)
 
+        if timestamp is not None:
+            # Timestamp the outgoing packets when a Timestamp source is provided
+            # and expose value to a software.
+            timestampbits        = len(timestamp)
+            self._timestamp_slot = CSRStatus(slotbits)
+            self._timestamp      = CSRStatus(timestampbits)
+
         self.submodules.ev = EventManager()
-        self.ev.done       = EventSourcePulse()
+        self.ev.done       = EventSourcePulse() if timestamp is None else EventSourceLevel()
         self.ev.finalize()
 
         # # #
 
+        start = Signal()
+
         # Command FIFO
-        fifo = stream.SyncFIFO([("slot", slotbits), ("length", lengthbits)], nslots)
-        self.submodules += fifo
+        cmd_fifo = stream.SyncFIFO([("slot", slotbits), ("length", lengthbits)], nslots)
+        self.submodules += cmd_fifo
         self.comb += [
-            fifo.sink.valid.eq(self._start.re),
-            fifo.sink.slot.eq(self._slot.storage),
-            fifo.sink.length.eq(self._length.storage),
-            self._ready.status.eq(fifo.sink.ready),
-            self._level.status.eq(fifo.level)
+            cmd_fifo.sink.valid.eq(self._start.re),
+            cmd_fifo.sink.slot.eq(self._slot.storage),
+            cmd_fifo.sink.length.eq(self._length.storage),
+            self._ready.status.eq(cmd_fifo.sink.ready),
+            self._level.status.eq(cmd_fifo.level)
         ]
 
+        # Status FIFO (Only added when Timestamping).
+        if timestamp is not None:
+            stat_fifo_layout = [("slot", slotbits), ("timestamp", timestampbits)]
+            stat_fifo = stream.SyncFIFO(stat_fifo_layout, nslots)
+            self.submodules += stat_fifo
+            self.comb += [
+                stat_fifo.source.ready.eq(self.ev.done.clear),
+                self._timestamp_slot.status.eq(stat_fifo.source.slot),
+                self._timestamp.status.eq(stat_fifo.source.timestamp),
+            ]
+
         # Length computation
         read_address = Signal(lengthbits)
         counter      = Signal(lengthbits)
@@ -183,12 +220,13 @@ class LiteEthMACSRAMReader(Module, AutoCSR):
         self.submodules.fsm = fsm = FSM(reset_state="IDLE")
         fsm.act("IDLE",
             NextValue(counter, 0),
-            If(fifo.source.valid,
-                read_address.eq(0),
+            If(cmd_fifo.source.valid,
+                start.eq(1),
                 NextState("SEND")
             )
         )
-        length_lsb = fifo.source.length[0:2]
+
+        length_lsb = cmd_fifo.source.length[0:2]
         if endianness == "big":
             self.comb += If(source.last,
                 Case(length_lsb, {
@@ -207,7 +245,7 @@ class LiteEthMACSRAMReader(Module, AutoCSR):
                 }))
         fsm.act("SEND",
             source.valid.eq(1),
-            source.last.eq(counter >= (fifo.source.length - 4)),
+            source.last.eq(counter >= (cmd_fifo.source.length - 4)),
             read_address.eq(counter),
             If(source.ready,
                 read_address.eq(counter + 4),
@@ -218,13 +256,21 @@ class LiteEthMACSRAMReader(Module, AutoCSR):
             )
         )
         fsm.act("END",
-            fifo.source.ready.eq(1),
             self.ev.done.trigger.eq(1),
+            cmd_fifo.source.ready.eq(1),
             NextState("IDLE")
         )
 
+        if timestamp is not None:
+            # Latch Timestamp on start of outgoing packet.
+            self.sync == If(start, stat_fifo.sink.timestamp.eq(timestamp))
+            self.comb += stat_fifo.sink.valid.eq(fsm.ongoing("IDLE")),
+            self.comb += stat_fifo.sink.slot.eq(cmd_fifo.source.slot)
+            # Trigger event when Status FIFO has contents.
+            self.comb += self.ev.done.trigger.eq(stat_fifo.source.valid),
+
         # Memory
-        rd_slot = fifo.source.slot
+        rd_slot = cmd_fifo.source.slot
         mems    = [None]*nslots
         ports   = [None]*nslots
         for n in range(nslots):
@@ -242,8 +288,8 @@ class LiteEthMACSRAMReader(Module, AutoCSR):
 # MAC SRAM -----------------------------------------------------------------------------------------
 
 class LiteEthMACSRAM(Module, AutoCSR):
-    def __init__(self, dw, depth, nrxslots, ntxslots, endianness):
-        self.submodules.writer = LiteEthMACSRAMWriter(dw, depth, nrxslots, endianness)
-        self.submodules.reader = LiteEthMACSRAMReader(dw, depth, ntxslots, endianness)
+    def __init__(self, dw, depth, nrxslots, ntxslots, endianness, timestamp=None):
+        self.submodules.writer = LiteEthMACSRAMWriter(dw, depth, nrxslots, endianness, timestamp)
+        self.submodules.reader = LiteEthMACSRAMReader(dw, depth, ntxslots, endianness, timestamp)
         self.submodules.ev     = SharedIRQ(self.writer.ev, self.reader.ev)
         self.sink, self.source = self.writer.sink, self.reader.source
